# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
from pprint import pformat
import logging
from requests import Session, adapters


logger = logging.getLogger(__name__)

DEFAULT_MAX_RETRIES = 1


def Requester(**kwargs):
    # Remote end closed connection without
    # response on successive Session requests maybe happen
    # see: https://github.com/psf/requests/issues/4784
    # and https://github.com/psf/requests/issues/4664
    session = Session()
    max_retries = kwargs.pop('max_retries', DEFAULT_MAX_RETRIES)
    adapter = adapters.HTTPAdapter(max_retries=max_retries)
    session.mount('http://', adapter)
    session.mount('https://', adapter)

    def send(method, url, **kw):
        logger.debug('%s: %s with parameters: %s',
                     method, url, pformat(kw))
        resp = session.request(method, url, **kwargs, **kw)
        logger.debug('Response: %s', resp)
        resp.raise_for_status()
        return resp

    return send



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""

from .item import Item
from .mix import ConfigurationMixIn, DeletionMixIn


class Credentials(Item):

    def get(self, id):
        for item in self.api_json(tree='credentials[id]')['credentials']:
            if item['id'] == id:
                return Credential(self.jenkins,
                                  f'{self.url}credential/{id}/')
        return None

    def create(self, xml):
        self.handle_req('POST', 'createCredentials',
                        headers=self.headers, data=xml)

    def __iter__(self):
        for item in self.api_json(tree='credentials[id]')['credentials']:
            yield Credential(self.jenkins,
                             f'{self.url}credential/{item["id"]}/')


class Credential(Item, ConfigurationMixIn, DeletionMixIn):
    pass



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import json
import time

import xml.etree.ElementTree as ET

from .item import Item


class PluginsManager(Item):

    def get(self, name):
        for plugin in self.api_json(tree='plugins[shortName]')['plugins']:
            if plugin['shortName'] == name:
                return Plugin(self.jenkins,
                              f'{self.url}plugin/{name}/')
        return None

    def install(self, *names, block=False):
        plugin_xml = ET.Element('jenkins')
        for name in names:
            if '@' not in name:
                name += '@latest'
            ET.SubElement(plugin_xml, 'install', {'plugin': name})
        self.handle_req('POST', 'installNecessaryPlugins',
                        headers=self.headers,
                        data=ET.tostring(plugin_xml))

        while block and not self.installation_done:
            time.sleep(2)

    def uninstall(self, *names):
        for name in names:
            self.handle_req('POST', f'plugin/{name}/doUninstall')

    def set_site(self, url):
        self.handle_req('POST', 'siteConfigure', params={'site': url})
        self.check_updates_server()

    def check_updates_server(self):
        self.handle_req('POST', 'checkUpdatesServer')

    @property
    def update_center(self):
        return UpdateCenter(self.jenkins, f'{self.jenkins.url}updateCenter/')

    @property
    def site(self):
        return self.update_center.site

    @property
    def restart_required(self):
        return self.update_center.restart_required

    @property
    def installation_done(self):
        return self.update_center.installation_done

    def set_proxy(self, name, port, *, username='',
                  password='', no_proxy='', test_url=''):
        data = {'name': name, 'port': port, 'userName': username,
                'password': password, 'noProxyHost': no_proxy,
                'testUrl': test_url}
        self.handle_req('POST', 'proxyConfigure', data={
                        'json': json.dumps(data)})

    def __iter__(self):
        for plugin in self.api_json(tree='plugins[shortName]')['plugins']:
            yield Plugin(self.jenkins,
                         f'{self.url}plugin/{plugin["shortName"]}/')


class Plugin(Item):
    def uninstall(self):
        self.handle_req('POST', 'doUninstall')


class UpdateCenter(Item):

    @property
    def installation_done(self):
        resp = self.handle_req('GET', 'installStatus')
        return all(job['installStatus'] != 'Pending'
                   for job in resp.json()['data']['jobs'])

    @property
    def restart_required(self):
        return self.api_json(tree='restartRequiredForCompletion').get(
            'restartRequiredForCompletion')

    @property
    def site(self):
        return self.api_json(tree='sites[url]')['sites'][0].get('url')



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import weakref
from importlib import import_module
from pathlib import PurePosixPath

from requests.exceptions import ConnectionError, HTTPError

from .__version__ import (__author__, __author_email__, __copyright__,
                          __description__, __license__, __title__, __url__,
                          __version__)
from .credential import Credentials
from .exceptions import ItemNotFoundError, AuthenticationError
from .item import Item
from .job import Folder
from .node import Nodes
from .plugin import PluginsManager
from .queue import Queue
from .requester import Requester
from .system import System
from .user import User, Users
from .view import Views

EMPTY_FOLDER_XML = '''<?xml version='1.0' encoding='UTF-8'?>
<com.cloudbees.hudson.plugins.folder.Folder/>'''


class Jenkins(Item):
    r'''Constructs  :class:`Jenkins <Jenkins>`.

    :param url: URL of Jenkins server, ``str``
    :param auth: (optional) Auth ``tuple`` to enable Basic/Digest/Custom HTTP Auth.
    :param token: (optional) Boolean, Create user token when initialize instance and
        revoke token once instance is destroied. useful when LDAP server refuse
        username and password used too much often. Defaults to ``False``.
    :param \*\*kwargs: other kwargs are same as `requests.Session.request <https://requests.readthedocs.io/en/latest/api/#requests.Session.request>`_

    Usage::

        >>> from api4jenkins import Jenkins
        >>> j = Jenkins('http://127.0.0.1:8080/', auth=('admin', 'admin'))
        >>> print(j)
        <Jenkins: http://127.0.0.1:8080/>
        >>> j.version
        '2.176.2'
    '''

    def __init__(self, url, **kwargs):
        token = kwargs.pop('token', None)
        self.send_req = Requester(**kwargs)
        super().__init__(self, url)
        self._crumb = None
        self._token = None
        self._auth = kwargs.get('auth', None)
        self.user = None
        if self._auth:
            self.user = User(self, f'{self.url}user/{self._auth[0]}/')
        if self._auth and token is True:
            self._token = self.user.generate_token()
            weakref.finalize(self.user, self.user.revoke_token,
                             self._token.uuid)
            kwargs['auth'] = (self._auth[0], self._token.value)
            self.send_req = Requester(**kwargs)

    def get_job(self, full_name):
        '''Get job by full name

        :param full_name: ``str``, full name of job
        :returns: Corresponding Job object or None

        Usage::

            >>> from api4jenkins import Jenkins
            >>> j = Jenkins('http://127.0.0.1:8080/', auth=('admin', 'admin'))
            >>> job = j.get_job('freestylejob')
            >>> print(job)
            <FreeStyleProject: http://127.0.0.1:8080/job/freestylejob/>
        '''
        folder, name = self._resolve_name(full_name)
        if folder.exists():
            return folder.get(name)

    def iter_jobs(self, depth=0):
        '''Iterate jobs with depth

        :param depth: ``int``, depth to iterate, default is 0
        :returns: iterator of jobs

        Usage::

            >>> from api4jenkins import Jenkins
            >>> j = Jenkins('http://127.0.0.1:8080/', auth=('admin', 'admin'))
            >>> for job in j.iter_jobs():
            ...     print(job)
            <FreeStyleProject: http://127.0.0.1:8080/job/freestylejob/>
            ...
        '''
        folder = Folder(self, self.url)
        yield from folder.iter(depth)

    def create_job(self, full_name, xml, recursive=False):
        '''Create new jenkins job with given xml configuration

        :param full_name: ``str``, full name of job
        :param xml: xml configuration string
        :param recursive: (optional) Boolean, recursively create folder if not existed

        Usage::

            >>> from api4jenkins import Jenkins
            >>> j = Jenkins('http://127.0.0.1:8080/', auth=('admin', 'admin'))
            >>> xml = """<?xml version='1.1' encoding='UTF-8'?>
            ... <project>
            ...   <builders>
            ...     <hudson.tasks.Shell>
            ...       <command>echo $JENKINS_VERSION</command>
            ...     </hudson.tasks.Shell>
            ...   </builders>
            ... </project>"""
            >>> j.create_job('freestylejob', xml)
            >>> job = j.get_job('freestylejob')
            >>> print(job)
            <FreeStyleProject: http://127.0.0.1:8080/job/freestylejob/>
        '''
        folder, name = self._resolve_name(full_name)
        if recursive and not folder.exists():
            self.create_job(folder.full_name, EMPTY_FOLDER_XML, recursive=recursive)
        return folder.create(name, xml)

    def copy_job(self, full_name, dest):
        '''Create job by copying other job, the source job and dest job are in
        same folder.

        :param full_name: full name of source job
        :param dest: name of new job

        Usage::

            >>> from api4jenkins import Jenkins
            >>> j = Jenkins('http://127.0.0.1:8080/', auth=('admin', 'admin'))
            >>> j.copy_job('folder/freestylejob', 'newjob')
            >>> j.get_job('folder/newjob')
            >>> print(job)
            <FreeStyleProject: http://127.0.0.1:8080/job/folder/job/newjob/>
        '''
        folder, name = self._resolve_name(full_name)
        return folder.copy(name, dest)

    def delete_job(self, full_name):
        '''Delete job

        :param full_name: ``str``, full name of job

        Usage::

            >>> from api4jenkins import Jenkins
            >>> j = Jenkins('http://127.0.0.1:8080/', auth=('admin', 'admin'))
            >>> job = j.get_job('freestylejob')
            >>> print(job)
            <FreeStyleProject: http://127.0.0.1:8080/job/freestylejob/>
            >>> j.delete_job('freestylejob')
            >>> job = j.get_job('freestylejob')
            >>> print(job)
            None
        '''
        job = self.get_job(full_name)
        if job:
            job.delete()

    def build_job(self, full_name, **params):
        '''Build job with/without params

        :param full_name: ``str``, full name of job
        :param params: parameters for building, support delay and remote token
        :returns: ``QueueItem``

        Usage::

            >>> from api4jenkins import Jenkins
            >>> j = Jenkins('http://127.0.0.1:8080/', auth=('admin', 'admin'))
            >>> item = j.build_job('freestylejob')
            >>> import time
            >>> while not item.get_build():
            ...      time.sleep(1)
            >>> build = item.get_build()
            >>> print(build)
            <FreeStyleBuild: http://127.0.0.1:8080/job/freestylejob/1/>
            >>> for line in build.progressive_output():
            ...     print(line)
            ...
        '''
        job = self.get_job(full_name)
        if job is None:
            raise ItemNotFoundError(f'No such job: {full_name}')
        return job.build(**params)

    def check_job_name(self, name):
        resp = self.handle_req('GET', 'checkJobName', params={'value': name})
        return 'is an unsafe character' in resp.text

    def validate_jenkinsfile(self, content):
        """validate Jenkinsfile, see
        https://www.jenkins.io/doc/book/pipeline/development/#linter

        Args:
            content (str): content of Jenkinsfile

        Returns:
            str: 'Jenkinsfile successfully validated.' if validate successful
            or error message
        """
        data = {'jenkinsfile': content}
        return self.handle_req(
            'POST', 'pipeline-model-converter/validate', data=data).text

    def _url2name(self, url):
        '''Covert job url to full name

        :param url: ``str``, url of job
        :returns: ``str``, full name of job
        '''
        if not url.startswith(self.url):
            raise ValueError(f'{url} is not in {self.url}')
        return url.replace(self.url, '/').replace('/job/', '/').strip('/')

    def _name2url(self, full_name):
        '''Covert job full name to url

        :param full_name: ``str``, full name of job
        :returns: ``str``, url of job
        '''
        if not full_name:
            return self.url
        full_name = full_name.strip('/').replace('/', '/job/')
        return f'{self.url}job/{full_name}/'

    def _resolve_name(self, full_name):
        '''Resolve folder and job name from full name'''
        if full_name.startswith(('http://', 'https://')):
            full_name = self._url2name(full_name)
        path = PurePosixPath(full_name)
        parent = str(path.parent) if path.parent.name else ''
        return Folder(self, self._name2url(parent)), path.name

    def exists(self):
        '''Check if Jenkins server is up

        :returns: Ture or False
        '''
        try:
            self.send_req('GET', self.url)
            return True
        except ConnectionError:
            return False
        except HTTPError as e:
            return e.response.status_code in [401, 403]

    @property
    def crumb(self):
        '''Crumb of Jenkins'''
        if self._crumb is None:
            try:
                _crumb = self.send_req(
                    'GET', f'{self.url}crumbIssuer/api/json').json()
                self._crumb = {_crumb['crumbRequestField']: _crumb['crumb']}
            except HTTPError as e:
                if e.response.status_code in [401, 403]:
                    raise AuthenticationError(
                        f'Invalid authorization for {self}') from e
                self._crumb = {}
        return self._crumb

    @property
    def system(self):
        '''An object for managing system operation.
        see :class:`System <api4jenkins.system.System>`'''
        return System(self, self.url)

    @property
    def plugins(self):
        '''An object for managing plugins.
        see :class:`PluginsManager <api4jenkins.plugin.PluginsManager>`'''
        return PluginsManager(self, f'{self.url}pluginManager/')

    @property
    def version(self):
        '''Version of Jenkins'''
        return self.handle_req('GET', '').headers['X-Jenkins']

    @property
    def credentials(self):
        '''An object for managing credentials.
        see :class:`Credentials <api4jenkins.credential.Credentials>`'''
        return Credentials(self,
                           f'{self.url}credentials/store/system/domain/_/')

    @property
    def views(self):
        '''An object for managing views of main window.
        see :class:`Views <api4jenkins.view.Views>`'''
        return Views(self)

    @property
    def nodes(self):
        '''An object for managing nodes.
        see :class:`Nodes <api4jenkins.node.Nodes>`'''
        return Nodes(self, f'{self.url}computer/')

    @property
    def queue(self):
        '''An object for managing build queue.
        see :class:`Queue <api4jenkins.queue.Queue>`'''
        return Queue(self, f'{self.url}queue/')

    @property
    def users(self):
        return Users(self, f'{self.url}asynchPeople/')

    @property
    def me(self):
        return self.user

    def __iter__(self):
        yield from self.iter_jobs()

    def __call__(self, depth):
        yield from self.iter_jobs(depth)

    def __getitem__(self, full_name):
        return self.get_job(full_name)


def _patch_to(module, cls, func=None):
    _module = import_module(module)
    if func:
        _class = getattr(_module, cls.__name__)
        setattr(_class, func.__name__, func)
    else:
        setattr(_module, cls.__name__, cls)



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import json

from .exceptions import ItemNotFoundError
from .item import Item, new_item
from .mix import ConfigurationMixIn, DeletionMixIn, RunScriptMixIn


class Nodes(Item):
    '''
    classdocs
    '''

    def create(self, name, **kwargs):
        node_setting = {
            'nodeDescription': '',
            'numExecutors': 1,
            'remoteFS': '/home/jenkins',
            'labelString': '',
            'mode': 'NORMAL',
            'retentionStrategy': {
                'stapler-class': 'hudson.slaves.RetentionStrategy$Always'
            },
            'nodeProperties': {'stapler-class-bag': 'true'},
            'launcher': {'stapler-class': 'hudson.slaves.JNLPLauncher'}
        }
        node_setting.update(kwargs)
        params = {
            'name': name,
            'type': 'hudson.slaves.DumbSlave$DescriptorImpl',
            'json': json.dumps(node_setting)
        }
        self.handle_req('POST', 'doCreateItem', data=params)

    def get(self, name):
        for item in self.api_json(tree='computer[displayName]')['computer']:
            if name == item['displayName']:
                item['url'] = f"{self.url}{item['displayName']}/"
                return self._new_instance_by_item(__name__, item)
        return None

    def iter_builds(self):
        builds = {}
        # iterate 'executors', 'oneOffExecutors' in order,
        # cause freestylebuild is in executors, and _class of workflowbuild in
        # executors is PlaceholderExecutable
        tree = ('computer[executors[currentExecutable[url]],'
                'oneOffExecutors[currentExecutable[url]]]')
        for computer in self.api_json(tree, 2)['computer']:
            _parse_builds(computer, builds)

        yield from _new_items(self.jenkins, builds)

    def iter_building_builds(self):
        yield from filter(lambda build: build.building, self.iter_builds())

    def __iter__(self):
        for item in self.api_json(tree='computer[displayName]')['computer']:
            item['url'] = f"{self.url}{item['displayName']}/"
            yield self._new_instance_by_item(__name__, item)

    def filter_node_by_label(self, *labels):
        for node in self:
            for label in node.api_json()['assignedLabels']:
                if label['name'] in labels:
                    yield node

    def filter_node_by_status(self, *, online):
        yield from filter(lambda node: online != node.offline, self)

# following two functions should be used in this module only


def _new_items(jenkins, builds):
    for url, class_name in builds.items():
        item = {'url': url, '_class': class_name}
        yield new_item(jenkins, 'api4jenkins.build', item)


def _parse_builds(data, builds):
    for kind in ['executors', 'oneOffExecutors']:
        for executor in data.get(kind):
            # in case of issue:
            # https://github.com/joelee2012/api4jenkins/issues/16
            execable = executor['currentExecutable']
            if not execable:
                continue
            if execable['_class'].endswith('PlaceholderExecutable'):
                execable['_class'] = 'org.jenkinsci.plugins.workflow.job.WorkflowRun'
            builds[execable['url']] = execable['_class']


class Node(Item, ConfigurationMixIn, DeletionMixIn, RunScriptMixIn):

    def enable(self):
        if self.offline:
            self.handle_req('POST', 'toggleOffline',
                            params={'offlineMessage': ''})

    def disable(self, msg=''):
        if not self.offline:
            self.handle_req('POST', 'toggleOffline',
                            params={'offlineMessage': msg})

    def iter_builds(self):
        builds = {}
        # iterate 'executors', 'oneOffExecutors' in order,
        # cause freestylebuild is in executors
        tree = ('executors[currentExecutable[url]],'
                'oneOffExecutors[currentExecutable[url]]')
        _parse_builds(self.api_json(tree, 2), builds)
        yield from _new_items(self.jenkins, builds)

    def __iter__(self):
        yield from self.iter_builds()

    def iter_building_builds(self):
        yield from filter(lambda build: build.building, self.iter_builds())


class MasterComputer(Node):
    def __init__(self, jenkins, url):
        # rename built-in node: https://www.jenkins.io/doc/upgrade-guide/2.319/
        name = 'master' if url.endswith('/master/') else 'built-in'
        super().__init__(jenkins, f'{jenkins.url}computer/({name})/')


class SlaveComputer(Node):
    pass


class KubernetesComputer(Node):

    def exists(self):
        try:
            self.handle_req('GET', '')
            return True
        except ItemNotFoundError:
            return False


class DockerComputer(Node):
    pass


class EC2Computer(Node):
    pass



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import re
from .item import Item
from .mix import ActionsMixIn


class Queue(Item):

    def get(self, id):
        for item in self.api_json(tree='items[id,url]')['items']:
            if item['id'] == int(id):
                return QueueItem(self.jenkins,
                                 f"{self.jenkins.url}{item['url']}")
        return None

    def cancel(self, id):
        self.handle_req('POST', 'cancelItem', params={
                        'id': id}, allow_redirects=False)

    def __iter__(self):
        for item in self.api_json(tree='items[url]')['items']:
            yield QueueItem(self.jenkins, f"{self.jenkins.url}{item['url']}")

# https://javadoc.jenkins.io/hudson/model/Queue.html#buildables
#  (enter) --> waitingList --+--> blockedProjects
#                            |        ^
#                            |        |
#                            |        v
#                            +--> buildables ---> pending ---> left
#                                     ^              |
#                                     |              |
#                                     +---(rarely)---+


class QueueItem(Item, ActionsMixIn):

    def __init__(self, jenkins, url):
        if not url.startswith('https') and jenkins.url.startswith('https'):
            url = re.sub(r'^http[s]', 'https', url)
        super().__init__(jenkins, url)
        self.id = int(self.url.split('/')[-2])
        self._build = None

    def get_job(self):
        if self._class.endswith('$BuildableItem'):
            return self.get_build().get_job()
        task = self.api_json(tree='task[url]')['task']
        return self._new_instance_by_item('api4jenkins.job', task)

    def get_build(self):
        if not self._build:
            _class = self._class
            # BlockedItem does not have build
            if _class.endswith('$LeftItem'):
                executable = self.api_json('executable[url]')['executable']
                self._build = self._new_instance_by_item(
                    'api4jenkins.build', executable)
            elif _class.endswith(('$BuildableItem', '$WaitingItem')):
                for build in self.jenkins.nodes.iter_builds():
                    # https://javadoc.jenkins.io/hudson/model/Run.html#getQueueId--
                    # https://javadoc.jenkins.io/hudson/model/Queue.Item.html#getId--
                    # ensure build exists, see https://github.com/joelee2012/api4jenkins/issues/49
                    if build.exists() and int(build.queue_id) == self.id:
                        self._build = build
                        break
        return self._build

    def cancel(self):
        self.jenkins.queue.cancel(self.id)

# due to item type is dynamic


class BuildableItem(QueueItem):
    pass


class BlockedItem(QueueItem):
    pass


class LeftItem(QueueItem):
    pass


class WaitingItem(QueueItem):
    pass



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
from .item import Item, camel, snake


class GetMixIn:

    def get(self, name):
        for item in self:
            if item.name == name:
                return item
        return None


class ResultBase:

    def __init__(self, raw):
        self.raw = raw

    def __getattr__(self, name):
        camel_name = camel(name)
        if camel_name in self.raw:
            return self.raw[camel_name]
        return super().__getattribute__(name)

    def __str__(self):
        return f'<{type(self).__name__}: {self.name}>'

    def __dir__(self):
        return super().__dir__() + [snake(k) for k in self.raw]


class TestReport(Item, GetMixIn):

    def __iter__(self):
        for suite in self.api_json()['suites']:
            yield TestSuite(suite)

    @property
    def suites(self):
        yield from self


class TestSuite(ResultBase, GetMixIn):

    def __iter__(self):
        for case in self.raw['cases']:
            yield TestCase(case)

    @property
    def cases(self):
        yield from self


class TestCase(ResultBase):
    pass


class CoverageReport(Item, GetMixIn):
    '''Access coverage report generated by `JaCoCo <https://plugins.jenkins.io/jacoco/>`_'''

    report_types = ['branchCoverage', 'classCoverage', 'complexityScore',
                    'instructionCoverage', 'lineCoverage', 'methodCoverage']

    def __getattr__(self, name):
        attr = camel(name)
        if attr not in self.report_types:
            raise AttributeError(
                f"'CoverageReport' object has no attribute '{name}'")
        return self.get(attr)

    def __iter__(self):
        for k, v in self.api_json().items():
            if k != '_class' and k != 'previousResult':
                v['name'] = k
                yield Coverage(v)

    def trends(self, count=2):
        def _resolve(data):
            if data['previousResult']:
                yield from _resolve(data['previousResult'])
            for k, v in data.items():
                if k != '_class' and k != 'previousResult':
                    v['name'] = k
                    yield Coverage(v)

        yield from _resolve(self.api_json(depth=count))


class Coverage(ResultBase):
    pass


class CoverageResult(Item, GetMixIn):
    '''Access coverage result generated by `Code Coverage API <https://plugins.jenkins.io/code-coverage-api/>`_'''

    def __iter__(self):
        for element in self.api_json(depth=1)['results']['elements']:
            yield CoverageElement(element)


class CoverageElement(ResultBase):
    pass


class CoverageTrends(Item, GetMixIn):
    def __iter__(self):
        for trend in self.api_json(depth=1)['trends']:
            trend['name'] = trend['buildName']
            yield CoverageTrend(trend)


class CoverageTrend(ResultBase):

    def __iter__(self):
        for element in self.raw['elements']:
            yield CoverageElement(element)



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
from collections import namedtuple

from .item import Item
from .mix import DeletionMixIn, DescriptionMixIn


class Users(Item):

    tree = 'users[user[id,absoluteUrl,fullName]]'

    def __iter__(self):
        for user in self.api_json(depth=2, tree=self.tree)['users']:
            yield User(self.jenkins, user['user']['absoluteUrl'])

    def get(self, id=None, full_name=None):
        for user in self.api_json(depth=2, tree=self.tree)['users']:
            if id == user['user']['id'] or full_name == user['user']['fullName']:
                return User(self.jenkins, user['user']['absoluteUrl'])
        return None


ApiToken = namedtuple('ApiToken', ['name', 'uuid', 'value'])


class User(Item, DeletionMixIn, DescriptionMixIn):

    def generate_token(self, name=''):
        entry = 'descriptorByName/jenkins.security.' \
                'ApiTokenProperty/generateNewToken'
        data = self.handle_req('POST', entry,
                               params={'newTokenName': name}).json()['data']
        return ApiToken(data['tokenName'],
                        data['tokenUuid'], data['tokenValue'])

    def revoke_token(self, uuid):
        entry = 'descriptorByName/jenkins.security.' \
                'ApiTokenProperty/revoke'
        return self.handle_req('POST', entry,
                               params={'tokenUuid': uuid})



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
__version__ = '0.0.1'
__title__ = 'api66jenkins'
__description__ = 'Jenkins Python Client'
__url__ = 'https://github.com/joelee2012/api4jenkins'
__author__ = 'sanmejie'
__author_email__ = 'sanmejie@163.cm'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2022 Joe Lee'
__documentation__ = 'https://api4jenkins.readthedocs.io'



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import re
import time

from .artifact import Artifact, save_response_to
from .input import PendingInputAction
from .item import Item
from .mix import DeletionMixIn, DescriptionMixIn, ActionsMixIn
from .report import CoverageReport, CoverageResult, CoverageTrends, TestReport


class Build(Item, DescriptionMixIn, DeletionMixIn, ActionsMixIn):

    def console_text(self, stream=False):
        with self.handle_req('GET', 'consoleText', stream=stream) as resp:
            for line in resp.iter_lines():
                yield line

    def progressive_output(self, html=False):
        url = 'logText/progressiveHtml' if html else 'logText/progressiveText'
        start = 0
        while True:
            resp = self.handle_req('GET', url, params={'start': start})
            time.sleep(1)
            if start == resp.headers.get('X-Text-Size'):
                continue
            yield resp.text
            if not resp.headers.get('X-More-Data'):
                break
            start = resp.headers['X-Text-Size']

    def stop(self):
        return self.handle_req('POST', 'stop', allow_redirects=False)

    def term(self):
        return self.handle_req('POST', 'term', allow_redirects=False)

    def kill(self):
        return self.handle_req('POST', 'kill', allow_redirects=False)

    def get_next_build(self):
        item = self.api_json(tree='nextBuild[url]')['nextBuild']
        if item:
            return self.__class__(self.jenkins, item['url'])
        return None

    def get_previous_build(self):
        item = self.api_json(tree='previousBuild[url]')['previousBuild']
        if item:
            return self.__class__(self.jenkins, item['url'])
        return None

    def get_job(self):
        '''get job of this build'''
        job_name = self.jenkins._url2name(re.sub(r'\w+[/]?$', '', self.url))
        return self.jenkins.get_job(job_name)

    def get_test_report(self):
        tr = TestReport(self.jenkins, f'{self.url}testReport/')
        return tr if tr.exists() else None

    def get_coverage_report(self):
        '''Access coverage report generated by `JaCoCo <https://plugins.jenkins.io/jacoco/>`_'''
        cr = CoverageReport(self.jenkins, f'{self.url}jacoco/')
        return cr if cr.exists() else None

    def get_coverage_result(self):
        '''Access coverage result generated by `Code Coverage API <https://plugins.jenkins.io/code-coverage-api/>`_'''
        cr = CoverageResult(self.jenkins, f'{self.url}coverage/result/')
        return cr if cr.exists() else None

    def get_coverage_trends(self):
        ct = CoverageTrends(self.jenkins, f'{self.url}coverage/trend/')
        return ct if ct.exists() else None


class WorkflowRun(Build):

    def get_pending_input(self):
        '''get current pending input step'''
        data = self.handle_req('GET', 'wfapi/describe').json()
        if not data['_links'].get('pendingInputActions'):
            return None
        action = self.handle_req('GET', 'wfapi/pendingInputActions').json()[0]
        return PendingInputAction(self.jenkins, action)

    def get_artifacts(self):
        artifacts = self.handle_req('GET', 'wfapi/artifacts').json()
        return [Artifact(self.jenkins, art) for art in artifacts]

    def save_artifacts(self, filename='archive.zip'):
        with self.handle_req('GET', 'artifact/*zip*/archive.zip') as resp:
            save_response_to(resp, filename)


class FreeStyleBuild(Build):
    pass


class MatrixBuild(Build):
    pass



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""

from .item import Item
from .mix import ConfigurationMixIn, DeletionMixIn, DescriptionMixIn


class Views(Item):
    '''
    classdocs
    '''

    def __init__(self, owner):
        '''
        Constructor
        '''
        self.owner = owner
        super().__init__(owner.jenkins, owner.url)

    def get(self, name):
        for item in self.api_json(tree='views[name,url]')['views']:
            if name == item['name']:
                return self._new_instance_by_item(__name__, item)
        return None

    def create(self, name, xml):
        self.handle_req('POST', 'createView', params={'name': name},
                        headers=self.headers, data=xml)

    def __iter__(self):
        for item in self.api_json(tree='views[name,url]')['views']:
            yield self._new_instance_by_item(__name__, item)


class View(Item, ConfigurationMixIn, DescriptionMixIn, DeletionMixIn):

    def get_job(self, name):
        for item in self.api_json(tree='jobs[name,url]')['jobs']:
            if name == item['name']:
                return self._new_instance_by_item('api4jenkins.job', item)
        return None

    def __iter__(self):
        for item in self.api_json(tree='jobs[name,url]')['jobs']:
            yield self._new_instance_by_item('api4jenkins.job', item)

    def include(self, name):
        self.handle_req('POST', 'addJobToView', params={'name': name})

    def exclude(self, name):
        self.handle_req('POST', 'removeJobFromView', params={'name': name})


class AllView(View):
    def __init__(self, jenkins, url):
        # name of all view for jenkins is 'all', but for folder is 'All'
        name = 'view/all' if jenkins.url == url else 'view/All'
        super().__init__(jenkins, url + name)


class MyView(View):
    pass


class ListView(View):
    pass


class Dashboard(View):
    pass


class NestedView(View):

    @property
    def views(self):
        return Views(self)


class SectionedView(View):
    pass



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import re
from importlib import import_module

from requests.exceptions import HTTPError

from .exceptions import (AuthenticationError, BadRequestError,
                         ItemNotFoundError, ServerError)


def camel(s):
    if s[0] == '_':
        return s
    first, *other = s.split('_')
    return first.lower() + ''.join(x.title() for x in other)


def _snake():
    pattern = re.compile(r'(?<!^)(?=[A-Z])')

    def func(name):
        return pattern.sub('_', name).lower()
    return func


snake = _snake()


def append_slash(url):
    return url if url[-1] == '/' else f'{url}/'


def _new_item():
    delimiter = re.compile(r'[.$]')

    def func(jenkins, module, item):
        class_name = delimiter.split(item['_class'])[-1]
        module = import_module(module)
        if not hasattr(module, class_name):
            msg = f'''{module} has no class {class_name} to describe
                  {item["url"]}, patch new class with api4jenkins._patch_to,
                  see: https://api4jenkins.readthedocs.io/en/latest/user/example.html#patch'''
            raise AttributeError(msg)
        _class = getattr(module, class_name)
        return _class(jenkins, item['url'])
    return func


new_item = _new_item()


class Item:
    '''
    classdocs
    '''
    headers = {'Content-Type': 'text/xml; charset=utf-8'}
    _dynamic_attrs = []

    def __init__(self, jenkins, url):
        self.jenkins = jenkins
        self.url = append_slash(url)

    def api_json(self, tree='', depth=0):
        params = {'depth': depth}
        if tree:
            params['tree'] = tree
        return self.handle_req('GET', 'api/json', params=params).json()

    def handle_req(self, method, entry, **kwargs):
        self._add_crumb(kwargs)
        if 'data' in kwargs and isinstance(kwargs['data'], str):
            kwargs['data'] = kwargs['data'].encode('utf-8')
        try:
            return self.jenkins.send_req(method, self.url + entry, **kwargs)
        except HTTPError as e:
            if e.response.status_code == 404:
                raise ItemNotFoundError(
                    f'Not found {entry} for item: {self}') from e
            if e.response.status_code == 401:
                raise AuthenticationError(
                    f'Invalid authorization for {self}') from e
            if e.response.status_code == 403:
                raise PermissionError(
                    f'No permission to {entry} for {self}') from e
            if e.response.status_code == 400:
                raise BadRequestError(e.response.headers['X-Error']) from e
            if e.response.status_code == 500:
                #                 import xml.etree.ElementTree as ET
                #                 tree = ET.fromstring(e.response.text)
                #                 stack_trace = tree.find(
                #                     './/div[@id="error-description"]/pre').text

                raise ServerError(e.response.text) from e
            raise

    def _add_crumb(self, kwargs):
        if self.jenkins.crumb:
            headers = kwargs.get('headers', {})
            headers.update(self.jenkins.crumb)
            kwargs['headers'] = headers

    def _new_instance_by_item(self, module, item):
        return new_item(self.jenkins, module, item)

    def exists(self):
        try:
            self.api_json(tree='_class')
            return True
        except ItemNotFoundError:
            return False

    @property
    def dynamic_attrs(self):
        if not self._dynamic_attrs:
            data = self.api_json()
            self.__class__._dynamic_attrs = \
                [snake(key) for key, val in data.items() if isinstance(
                    val, (int, str, bool, type(None)))]
        return self._dynamic_attrs

    def __eq__(self, other):
        return type(self) is type(other) and self.url == other.url

    def __str__(self):
        return f'<{type(self).__name__}: {self.url}>'

    def __getattr__(self, name):
        if name in self.dynamic_attrs:
            attr = camel(name)
            return self.api_json(tree=attr)[attr]
        return super().__getattribute__(name)



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import json

from .item import Item
from .mix import RawJsonMixIn


class PendingInputAction(RawJsonMixIn, Item):
    ''' this class implement functionality to process
    `input step <https://www.jenkins.io/doc/pipeline/steps/pipeline-input-step/>`_
    '''

    def __init__(self, jenkins, raw):
        super().__init__(
            jenkins, f"{jenkins.url}{raw['abortUrl'].rstrip('abort')}")
        self.raw = raw
        self.raw['_class'] = 'PendingInputAction'

    def abort(self):
        '''submit `input step <https://www.jenkins.io/doc/pipeline/steps/pipeline-input-step/>`_'''
        return self.handle_req('POST', 'abort')

    def submit(self, **params):
        '''submit `input step <https://www.jenkins.io/doc/pipeline/steps/pipeline-input-step/>`_

        - for input requires parametes:
            - if submit without parameters, it will use default value of parameters
            - if submit with wrong parameters, exception raised
        - for input does not requires parameters, but submit with paramters, exception raised
        '''
        if params:
            args = [a['name'] for a in self.raw['inputs']]
            params_keys = list(params.keys())
            if not args:
                raise TypeError(
                    f'input takes 0 argument, but got {params_keys}')
            if not all(k in args for k in params_keys):
                raise TypeError(
                    f'input takes arguments: {args}, but got {params_keys}')
            params = [{'name': k, 'value': v} for k, v in params.items()]
            data = {'proceed': self.raw['proceedText'],
                    'json': json.dumps({'parameter': params})}
            return self.handle_req('POST', 'submit', data=data)
        return self.handle_req('POST', 'proceedEmpty')



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""

class ItemNotFoundError(Exception):
    pass


class AuthenticationError(Exception):
    pass


class ItemExistsError(Exception):
    pass


class UnsafeCharacterError(Exception):
    pass


class BadRequestError(Exception):
    pass


class ServerError(Exception):
    pass



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""

# pylint: disable=no-member

from collections import namedtuple


class DeletionMixIn:

    def delete(self):
        self.handle_req('POST', 'doDelete', allow_redirects=False)


class ConfigurationMixIn:

    def configure(self, xml=None):
        if not xml:
            return self.handle_req('GET', 'config.xml').text
        return self.handle_req('POST', 'config.xml',
                               headers=self.headers, data=xml)

    @property
    def name(self):
        return self.url.split('/')[-2]


class DescriptionMixIn:

    def set_description(self, text):
        self.handle_req('POST', 'submitDescription',
                        params={'description': text})


class RunScriptMixIn:

    def run_script(self, script):
        return self.handle_req('POST', 'scriptText',
                               data={'script': script}).text


class EnableMixIn:

    def enable(self):
        return self.handle_req('POST', 'enable')

    def disable(self):
        return self.handle_req('POST', 'disable')


class RawJsonMixIn:

    def api_json(self, tree='', depth=0):
        return self.raw


Parameter = namedtuple('Parameter', ['class_name', 'name', 'value'])


class ActionsMixIn:

    def get_parameters(self):
        parameters = []
        for action in self.api_json()['actions']:
            if 'parameters' in action:
                parameters.extend(Parameter(raw['_class'], raw['name'], raw.get(
                    'value', '')) for raw in action['parameters'])
                break
        return parameters

    def get_causes(self):
        return next((action['causes'] for action in self.api_json()['actions'] if 'causes' in action), [])



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import json
from functools import partial

from .artifact import save_response_to
from .item import Item, snake
from .mix import RunScriptMixIn


class System(Item, RunScriptMixIn):

    def __init__(self, jenkins, url):
        '''
        see: https://support.cloudbees.com/hc/en-us/articles/216118748-How-to-Start-Stop-or-Restart-your-Instance-
        '''
        super().__init__(jenkins, url)

        def _post(entry):
            return self.handle_req('POST', entry, allow_redirects=False)

        for entry in ['restart', 'safeRestart', 'exit',
                      'safeExit', 'quietDown', 'cancelQuietDown']:
            setattr(self, snake(entry), partial(_post, entry))

    def reload_jcasc(self):
        return self.handle_req('POST', 'configuration-as-code/reload')

    def export_jcasc(self, filename='jenkins.yaml'):
        with self.handle_req('POST', 'configuration-as-code/export') as resp:
            save_response_to(resp, filename)

    def apply_jcasc(self, new):
        params = {"newSource": new}
        resp = self.handle_req(
            'POST', 'configuration-as-code/checkNewSource', params=params)
        if resp.text.startswith('<div class=error>'):
            raise ValueError(resp.text)
        d = {'json': json.dumps(params),
             'replace': 'Apply new configuration'}
        return self.handle_req('POST', 'configuration-as-code/replace', data=d)

    def decrypt_secret(self, text):
        cmd = f'println(hudson.util.Secret.decrypt("{text}"))'
        return self.run_script(cmd)



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
import json
import xml.etree.ElementTree as ET
from functools import partial
from pathlib import PurePosixPath
from urllib.parse import unquote_plus

from .credential import Credentials
from .item import Item, append_slash, snake
from .mix import (ConfigurationMixIn, DeletionMixIn, DescriptionMixIn,
                  EnableMixIn)
from .queue import QueueItem
from .view import Views


class Job(Item, ConfigurationMixIn, DescriptionMixIn, DeletionMixIn):

    def move(self, path):
        path = path.strip('/')
        params = {'destination': f'/{path}',
                  'json': json.dumps({'destination': f'/{path}'})}
        resp = self.handle_req('POST', 'move/move',
                               data=params, allow_redirects=False)
        self.url = resp.headers['Location']
        return resp

    def rename(self, name):
        resp = self.handle_req('POST', 'confirmRename',
                               params={'newName': name},
                               allow_redirects=False)
        self.url = append_slash(resp.headers['Location'])
        return resp

    def duplicate(self, path, recursive=False):
        self.jenkins.create_job(path, self.configure(), recursive=recursive)

    @property
    def parent(self):
        path = PurePosixPath(self.full_name)
        if path.parent.name == '':
            return self.jenkins
        return self.jenkins.get_job(str(path.parent))

    @property
    def name(self):
        return self.full_name.split('/')[-1]

    @property
    def full_name(self):
        return unquote_plus(self.jenkins._url2name(self.url))

    @property
    def full_display_name(self):
        return unquote_plus(self.full_name.replace('/', ' » '))


class Folder(Job):

    def create(self, name, xml):
        return self.handle_req('POST', 'createItem', params={'name': name},
                               headers=self.headers, data=xml)

    def get(self, name):
        for item in self.api_json(tree='jobs[name,url]')['jobs']:
            if name == item['name']:
                return self._new_instance_by_item(__name__, item)
        return None

    def iter(self, depth=0):
        query = 'jobs[url]'
        query_format = 'jobs[url,%s]'
        for _ in range(int(depth)):
            query = query_format % query

        def _resolve(item):
            yield self._new_instance_by_item(__name__, item)
            jobs = item.get('jobs')
            if jobs:
                for job in jobs:
                    yield from _resolve(job)

        for item in self.api_json(tree=query)['jobs']:
            yield from _resolve(item)

    def copy(self, src, dest):
        params = {'name': dest, 'mode': 'copy', 'from': src}
        return self.handle_req('POST', 'createItem', params=params,
                               allow_redirects=False)

    def reload(self):
        return self.handle_req('POST', 'reload')

    @property
    def views(self):
        return Views(self)

    @property
    def credentials(self):
        return Credentials(self.jenkins,
                           f'{self.url}credentials/store/folder/domain/_/')

    def __iter__(self):
        yield from self.iter()

    def __call__(self, depth):
        yield from self.iter(depth)

    def __getitem__(self, name):
        return self.get(name)


class WorkflowMultiBranchProject(Folder, EnableMixIn):

    def scan(self, delay=0):
        return self.handle_req('POST', 'build', params={'delay': delay})

    def get_scan_log(self, stream=False):
        with self.handle_req('GET', 'indexing/consoleText',
                             stream=stream) as resp:
            yield from resp.iter_lines()

    @property
    def buildable(self):
        return ET.XML(self.configure()).find('disabled').text == 'false'


class OrganizationFolder(WorkflowMultiBranchProject):
    pass


class Project(Job, EnableMixIn):

    def __init__(self, jenkins, url):
        super().__init__(jenkins, url)

        def _get_build_by_key(key):
            item = self.api_json(tree=f'{key}[url]')[key]
            if item is None:
                return None
            return self._new_instance_by_item('api4jenkins.build', item)

        for key in ['firstBuild', 'lastBuild', 'lastCompletedBuild',
                    'lastFailedBuild', 'lastStableBuild', 'lastUnstableBuild',
                    'lastSuccessfulBuild', 'lastUnsuccessfulBuild']:
            setattr(self, snake(f'get_{key}'),
                    partial(_get_build_by_key, key))

    def build(self, **params):
        """"""
        reserved = ['token', 'delay']
        reserved_tow=["custom"]
        if not params or all(k in reserved for k in params): #检查传入参数是否包含关键字，或者没有参数
            entry = 'build'
        elif all(k in reserved_tow for k in params):  #检查k是否包括关键字custom，拿v ，**params 相当于将k=v 变为{k:v}转换后也是字典
            entry = 'buildWithParameters'
            if isinstance(params,dict):  #判断是不是字典
                params=params.get(reserved_tow[0])
        else:
            entry = 'buildWithParameters'
        resp = self.handle_req('POST', entry, params=params)
        return QueueItem(self.jenkins, resp.headers['Location'])

    def get_build(self, number):
        for item in self.api_json(tree='builds[number,url]')['builds']:
            if int(number) == int(item['number']):
                return self._new_instance_by_item('api4jenkins.build', item)
        return None

    def iter_builds(self):
        yield from self

    def iter_all_builds(self):
        for item in self.api_json(tree='allBuilds[number,url]')['allBuilds']:
            yield self._new_instance_by_item('api4jenkins.build', item)

    def set_next_build_number(self, number):
        self.handle_req('POST', 'nextbuildnumber/submit',
                        params={'nextBuildNumber': number})

    def get_parameters(self):
        params = []
        for p in self.api_json()['property']:
            if 'parameterDefinitions' in p:
                params = p['parameterDefinitions']
        return params

    @property
    def building(self):
        builds = self.api_json(tree='builds[building]')['builds']
        return any(b['building'] for b in builds)

    def __iter__(self):
        for item in self.api_json(tree='builds[number,url]')['builds']:
            yield self._new_instance_by_item('api4jenkins.build', item)

    def __getitem__(self, number):
        return self.get_build(number)

    def filter_builds_by_result(self, *, result):
        """filter build by build results, avaliable results are:
        'SUCCESS', 'UNSTABLE', 'FAILURE', 'NOT_BUILT', 'ABORTED'
        see: https://javadoc.jenkins-ci.org/hudson/model/Result.html
        """
        expect = ['SUCCESS', 'UNSTABLE', 'FAILURE', 'NOT_BUILT', 'ABORTED']
        if result not in expect:
            raise ValueError(f'Expect one of {expect}')
        yield from filter(lambda build: build.result == result, self)


class WorkflowJob(Project):
    pass


class MatrixProject(Project):
    pass


class FreeStyleProject(Project):
    pass


class MavenModuleSet(Project):
    pass


class ExternalJob(Project):
    pass


class MultiJobProject(Project):
    pass


class IvyModuleSet(Project):
    pass


class BitbucketSCMNavigator(Project):
    pass


class GitHubSCMNavigator(Project):
    pass


class PipelineMultiBranchDefaultsProject(Project):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['api66jenkins']

package_data = \
{'': ['*']}

install_requires = \
['api4jenkins>=1.12,<2.0', 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'api66jenkins',
    'version': '0.0.7',
    'description': 'Python3 client library for Jenkins API. Adding a keyword from api4jenkins  add job.build [custom] parameter',
    'long_description': '\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/api4jenkins)\n![PyPI - Wheel](https://img.shields.io/pypi/wheel/api4jenkins)\n[![Documentation Status](https://readthedocs.org/projects/api4jenkins/badge/?version=latest)](https://api4jenkins.readthedocs.io/en/latest/?badge=latest)\n![GitHub](https://img.shields.io/github/license/joelee2012/api4jenkins)\n\n\n# Jenkins Python Client\n\n[Python3](https://www.python.org/) client library for [Jenkins API](https://wiki.jenkins.io/display/JENKINS/Remote+access+API). \n\ndispose api4jenkins job.  Build could not pass dictionary format。\n\nThanks to the original authors of api4jenkins for providing such a nice package to manipulate Jenkins\n\n# Features\n\n- Object oriented, each Jenkins item has corresponding class, easy to use and extend\n- Base on `api/json`, easy to query/filter attribute of item\n- Setup relationship between class just like Jenkins item\n- Support api for almost every Jenkins item\n- Pythonic\n- Test with latest Jenkins LTS\n\n# Installation\n\n```bash\npython3 -m pip install api66jenkins\n```\n\n# Quick start\n\n```python\n>>> from api66jenkins import Jenkins\n>>> client = Jenkins(\'http://127.0.0.1:8080/\', auth=(\'admin\', \'admin\'))\n>>> client.version\n\'2.176.2\'\n>>> xml = """<?xml version=\'1.1\' encoding=\'UTF-8\'?>\n... <project>\n...   <builders>\n...     <hudson.tasks.Shell>\n...       <command>echo $JENKINS_VERSION</command>\n...     </hudson.tasks.Shell>\n...   </builders>\n... </project>"""\n>>> client.create_job(\'path/to/job\', xml)\n>>> import time\n>>> item = client.build_job(\'path/to/job\')\n>>> while not item.get_build():\n...      time.sleep(1)\n>>> build = item.get_build()\n>>> for line in build.progressive_output():\n...     print(line)\n...\nStarted by user admin\nRunning as SYSTEM\nBuilding in workspace /var/jenkins_home/workspace/freestylejob\n[freestylejob] $ /bin/sh -xe /tmp/jenkins2989549474028065940.sh\n+ echo $JENKINS_VERSION\n2.176.2\nFinished: SUCCESS\n>>> build.building\nFalse\n>>> build.result\n\'SUCCESS\'\n```\n\n# Documentation\nUser Guide and API Reference is available on [Read the Docs](https://api4jenkins.readthedocs.io/)\n\n',
    'author': 'sanmejie',
    'author_email': 'liyapong@foxmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# encoding: utf-8
"""
Copyright (c) [2022] [sanmejie]
[Software Name] is licensed under the Mulan PSL v1.
You can use this software according to the terms and conditions of the Mulan PSL v1.
You may obtain a copy of Mulan PSL v1 at:
    http://license.coscl.org.cn/MulanPSL
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
PURPOSE.
See the Mulan PSL v1 for more details.
"""
from .item import Item
from .mix import RawJsonMixIn


class Artifact(RawJsonMixIn, Item):

    def __init__(self, jenkins, raw):
        super().__init__(
            jenkins, f"{jenkins.url}{raw['url'][1:]}")
        # remove trailing slash
        self.url = self.url[:-1]
        self.raw = raw
        self.raw['_class'] = 'Artifact'

    def save(self, filename=None):
        if not filename:
            filename = self.name
        with self.handle_req('GET', '') as resp:
            save_response_to(resp, filename)


def save_response_to(response, filename):
    with open(filename, 'wb') as fd:
        for chunk in response.iter_content(chunk_size=128):
            fd.write(chunk)
